import pino, { BaseLogger } from 'pino';
import path from 'path';
import httpPino from 'pino-http';

export class LoggingService {
  static instance: LoggingService;
  logsPath: string = path.join(process.cwd(), 'logs');
  logger!: BaseLogger;
  level: string = process.env.LOG_LEVEL || 'info';

  /*
  *   1 - fatal
      2 - error
      3 - warn
      4 - info
      5 - debug
      6 - trace
  */ 
  constructor() {
    if ((!LoggingService.instance)) {
      LoggingService.instance = this;
      this.logger = pino({
        level: this.level,
        timestamp: pino.stdTimeFunctions.isoTime,
      }, this.transports);
    }
    return LoggingService.instance;
  }

  get commonRotationOptions() : any {
    return {
      interval: '1d',
      maxFiles: 10,
      path: this.logsPath,
      size: '10M',
      maxSize: '100M',
    };
  }

  get transports() {
    return pino.transport({
      targets: [
        {
          target: 'pino-rotating-file-stream',
          level: this.level,
          options: { 
            filename: 'app.log',
            ...this.commonRotationOptions
          },
        },
        {
          target: 'pino-pretty',
          level: this.level,
          options: {
            sync: true,
            colorized: true,
          }
        },
      ]
    });
  }

  get httpTransports() : any {
    return pino.transport({
      targets: [
        {
          target: 'pino-rotating-file-stream',
          level: this.level,
          options: { 
            filename: 'http.log',
            ...this.commonRotationOptions
          },
        },
      ],
    });
  }

  middleware(): any {
    return httpPino({ 
      logger:  pino({
        level: this.level,
        timestamp: pino.stdTimeFunctions.isoTime,
      }, this.httpTransports) });
  }

  debug(message: string, data?: any) {
    this.logger.debug(this._getMessageWidthObject(message, data));
  }

  info(message: string, data?: any) {
    this.logger.info(this._getMessageWidthObject(message, data));
  }

  warn(message: string, data?: any) {
    this.logger.warn(this._getMessageWidthObject(message, data));
  }

  error(error: any, message?: string) {
    this.logger.error(error, message);
  }

  fatal(message: string, data?: any) {
    this.logger.fatal(this._getMessageWidthObject(message, data));
  }


  trace(message: string, data?: any) {
    this.logger.trace(this._getMessageWidthObject(message, data));
  }

  object(message: any) {
    this.logger.info(JSON.stringify(message, null, 2));
  }

  _getMessageWidthObject(message: string, data?: any) {
    if (!data) {
      return message;
    }
    return `${message}\n${this._getStringObject(data)}`;
  }

  _getStringObject(data: any) {
    try {
      return JSON.stringify(data, null, 2);
    } catch (error) {
      return data;
    }
  }

}