import { PlayerMoveSide } from "../../constants";
import { printLine, wait } from "../../../common/utilities";
import { AbstractPlayer } from "../../entities/player/AbstractPlayer";
import { Board } from "../Board";
import { PlayerMove } from "../PlayerMove";
import { SimulatedBoard } from "../../SimulatedBoard";
import { Tile } from "../Tile";
import { PRNG } from "seedrandom";

export class PlayerAI extends AbstractPlayer {
  constructor(name: string, private rng: PRNG) {
    super(name);
  }

  async makeMove(board: Board): Promise<PlayerMove | null> {
    const rndWait = Math.floor(Math.random() * 1000) + 1000;
    await wait(rndWait); // Simulate thinking time
    if (board.tiles.length === 0) {
      printLine('playing the first tile');
      const highestPair = this.getHighestPair();
      if (highestPair !== null) {
        return new PlayerMove(highestPair, PlayerMoveSide.BOTH, this.id);
      }
      const maxTile = this.getMaxTile();
      if (maxTile !== null) {
        return new PlayerMove(maxTile, PlayerMoveSide.BOTH, this.id);        
      }
    }
    // Analyze the game state
    // Return the best move based on strategy
    return this.chooseTileGreed(board);
  }

  async chooseTile(board: Board): Promise<Tile> {
    const randomWait = Math.floor((Math.random() * 1000) + 500);
    await wait(randomWait); // Simulate thinking time
    const randomIndex = Math.floor(this.rng() * board.boneyard.length);
    const tile = board.boneyard.splice(randomIndex, 1)[0];
    this.hand.push(tile);
    printLine(`${this.name} has chosen a tile`);
    return tile;
  }

  getMaxTile(): Tile | null {
    if (this.hand.length === 0) {
      return null;
    }

    let maxTile: Tile | null = null;
    this.hand.forEach(tile => {
      if (tile.count > (maxTile?.count ?? 0)) {
        maxTile = tile;
      }
    });
    return maxTile;
  } 

  chooseTileGreed(board: Board): PlayerMove | null { // greed algorithm
    let bestMove: PlayerMove |null = null;
    let bestTileScore: number = -1;
    const validMoves: PlayerMove[] = board.getValidMoves(this);

    validMoves.forEach(move => {
      const { tile } = move;
      const tileScore = tile.pips[0] + tile.pips[1];
      if (tileScore > bestTileScore) {
        bestMove = move;
        bestTileScore = tileScore;
      }
    });
    return bestMove;
  }

  chooseTileRandom(board: Board): Tile | null { // random algorithm
    const validTiles: Tile[] = this.hand.filter(tile => board.isValidMove(tile, null, this));
    return validTiles[Math.floor(this.rng() * validTiles.length)];
  }

  chooseTileMinMax(board: Board, depth: number = 3): Tile | null { // minmax algorithm
    const bestMove: Tile | null = null;
    let bestScore: number = -Infinity;
    const validMoves: PlayerMove[] = board.getValidMoves(this);

    validMoves.forEach(move => {
      const simulatedBoard = new SimulatedBoard([ ...board.tiles ], this.rng);
      simulatedBoard.play(move);
      const score = this.minmax(simulatedBoard, depth - 1, false);
    });

    return null;
  }

  minmax(simulatedBoard: SimulatedBoard, depth: number, isMaximizing: boolean): number {
    if (depth === 0 || simulatedBoard.isGameOver) {
      return simulatedBoard.evaluate();
    }

    if (isMaximizing) {
      let maxEval = -Infinity;
      const validMoves = simulatedBoard.getValidMoves(this);
      validMoves.forEach((move: PlayerMove) => {
        const newSimulatedBoard = new SimulatedBoard([ ...simulatedBoard.tiles ], this.rng);
        newSimulatedBoard.play(move);
        const evaluation: number = this.minmax(newSimulatedBoard, depth - 1, false);
        maxEval = Math.max(maxEval, evaluation);
      });
      return maxEval;
    }
     else {
      let minEval = Infinity;
      const validMoves = simulatedBoard.getValidMoves(this);
      validMoves.forEach((move: PlayerMove) => {
        const newSimulatedBoard = new SimulatedBoard([ ...simulatedBoard.tiles ], this.rng);
        newSimulatedBoard.play(move);
        const evaluation: number = this.minmax(newSimulatedBoard, depth - 1, true);
        minEval = Math.min(minEval, evaluation);
      });
      return minEval;
     }

    // if (isMaximizing) {
    //   let bestScore = -Infinity;
    //   boardTiles.forEach(tile => {
    //     const newBoardTiles = [ ...boardTiles ];
    //     newBoardTiles.push(tile);
    //     const score = this.minmax(newBoardTiles, depth - 1, false);
    //     bestScore = Math.max(score, bestScore);
    //   });
    //   return bestScore;
    // } else {
    //   let bestScore = Infinity;
    //   boardTiles.forEach(tile => {
    //     const newBoardTiles = [ ...boardTiles ];
    //     newBoardTiles.push(tile);
    //     const score = this.minmax(newBoardTiles, depth - 1, true);
    //     bestScore = Math.min(score, bestScore);
    //   });
      // return bestScore;
    // }
  }

  evaluateBoard(board: Board): number {
    // Custom heuristic to evaluate the board state
    return board.tiles.length; // Simplistic example
  }
}