import { SessionCreationError } from "../../common/errors/SessionCreationError";
import { SessionNotFoundError } from "../../common/errors/SessionNotFoundError";
import { wait, whileNotUndefined } from "../../common/utilities";
import { NetworkPlayer } from "../../game/entities/player/NetworkPlayer";
import { MatchSession } from "../../game/MatchSession";
import { PlayerNotificationService } from "./PlayerNotificationService";
import { matchSessionAdapter } from "../db/DbAdapter";
import { DbMatchSession } from "../db/interfaces";
import { MatchSessionMongoManager } from "../db/mongo/MatchSessionMongoManager";
import { SessionManager } from "../managers/SessionManager";
import { ServiceBase } from "./ServiceBase";
import { SocketIoService } from "./SocketIoService";

export class SessionService extends ServiceBase{
  private dbManager: MatchSessionMongoManager = new MatchSessionMongoManager();
  private sessionManager: SessionManager = new SessionManager();
  private notifyService = new PlayerNotificationService();  
  constructor() {
    super()
  }
  
  public async createSession(user: any, sessionName: string, seed: string ): Promise<string> {
    let socketClient;
    try {
      socketClient = await whileNotUndefined(() => SocketIoService.getClient(user._id));
    } catch (error) {
      throw new SessionCreationError();
    }
    const player =  new NetworkPlayer(user._id, user.username, socketClient.socketId);
    const session = new MatchSession(player, sessionName, seed);
    const dbSessionId = await this.dbManager.create(matchSessionAdapter(session));
    if (dbSessionId === undefined) {
      throw new SessionCreationError();
    }
    session.id = dbSessionId.toString();
    socketClient.sessionId = session.id;
    this.sessionManager.setSession(session);    
    this.notifyService.notifyMatchState(session);
    this.notifyService.notifyPlayersState(session.players);
    return session.id;
  }

  public async joinSession(user: any, sessionId: string): Promise<void> {
    const session: MatchSession | undefined = this.sessionManager.getSession(sessionId);
    if (session === undefined) {
      throw new SessionNotFoundError();
    }    let socketClient;
    try {
      socketClient = await whileNotUndefined(() => SocketIoService.getClient(user._id));
    } catch (error) {
      throw new SessionCreationError();
    }
    const player =  new NetworkPlayer(user._id, user.name, socketClient.socketId);
    session.addPlayer(player);
    socketClient.sessionId = session.id;
    this.dbManager.replaceOne({id: session.id}, matchSessionAdapter(session));
  }

  public listSessions(): Promise<DbMatchSession[]> {
    return this.dbManager.listByFilter({});
  }

  public updateSocketId(sessionId: string, userId: string, socketId: string): any {
    this.sessionManager.updateSocketId(sessionId, userId, socketId);
  }

  setPlayerReady(data: any): any {
    const { userId, sessionId } = data;
    const session: MatchSession | undefined = this.sessionManager.getSession(sessionId);
    if (session !== undefined) {
      session.setPlayerReady(userId)
      this.notifyService.notifyMatchState(session);
      this.notifyService.notifyPlayersState(session.players);
    }
  }

  startSession(data: any): any {
    const sessionId: string = data.sessionId;
    const seed: string | undefined = data.seed;
    const session: MatchSession | undefined = this.sessionManager.getSession(sessionId);

    if (session === undefined) {
      return ({
        status: 'error',
        message: 'Session not found'
      });
    } else if (session.matchInProgress) {
      return {
        status: 'error',
        message: 'Game already in progress'
      };
    } else {
      const missingHumans = session.maxPlayers - session.numPlayers;
      for (let i = 0; i < missingHumans; i++) {
        session.addPlayer(session.createPlayerAI(i));
      }
      session.start();
      return {
        status: 'ok'
      };
    }
  }

  // public updateSession(session: MatchSession): any {
  //   this.dbManager.replaceOne({id: session.id}, matchSessionAdapter(session));
  // }
}